use core::str;
use std::{env::args_os, fmt::Display, fs, process::exit};

use rand_pcg::Pcg64Mcg;
use sort_machine::parse;
use tasks::Task;
use unicode_segmentation::UnicodeSegmentation;

mod tasks;

const CORE_COUNT: usize = 2048;
const INPUT_COUNT: usize = 1;
const MAX_INSTS: usize = 256;

type Rng = Pcg64Mcg;

fn main() {
    let args = args_os().collect::<Box<[_]>>();
    let [_, input, _correct, program] = args.as_ref() else {
        eprintln!("Usage: [input] [correct output] [contestant output]");
        exit(1);
    };

    let input = fs::read_to_string(input).expect("input file should exist");
    let program = fs::read(program).expect("output file should exist");

    let task = Task::from_input(&input).expect("input file should be one of the inputs");

    let Ok(program) = str::from_utf8(&program) else {
        verdict(false, "Program obsahuje znaky mimo kódování UTF-8");
    };

    let program = match parse(program, INPUT_COUNT) {
        Ok(insts) => insts,
        Err(errors) => {
            let first = errors.into_iter().next().unwrap();
            verdict(false, format_args!("Chyba: {first}"));
        }
    };

    if program.len() > MAX_INSTS {
        verdict(false, "Program je moc dlouhý");
    }

    if task.judge(program) {
        verdict(true, SUCCESS);
    } else {
        verdict(false, WRONG);
    }
}

const SUCCESS: &str = "translate:success";
const WRONG: &str = "translate:wrong";

fn verdict(correct: bool, message: impl Display) -> ! {
    println!("{}", if correct { 1 } else { 0 });
    eprintln!("{}", limit_length(&message.to_string()));
    exit(0);
}

const MAX_CHARS: usize = 256;

fn limit_length(message: &str) -> String {
    let mut result = String::new();
    let mut chars = 0usize;

    for grapheme in message.graphemes(true) {
        let grapheme_length = grapheme.chars().count();

        if chars + grapheme_length > MAX_CHARS {
            result.push('…');
            return result;
        }

        result.push_str(grapheme);
        chars += grapheme_length;
    }

    result
}
