use std::{iter::repeat, ops::RangeInclusive};

use rand::{distributions::Uniform, Rng as _, SeedableRng};
use rayon::iter::{IndexedParallelIterator, IntoParallelRefIterator, ParallelIterator};
use sort_machine::{Instruction, Machine, Value};

use crate::{Rng, CORE_COUNT};

const SEED_BASE: u64 = 0x566dccb398a35e90;

fn judge_single(input: Box<[Value]>, program: &[Instruction]) -> bool {
    let expected = *input.iter().min().unwrap();
    let mut machine = Machine::new(CORE_COUNT, Box::new([input]));

    for inst in program {
        machine.exec(*inst);
    }

    machine.output()[0] == expected
}

pub fn judge(program: Vec<Instruction>) -> bool {
    input_gens()
        .par_iter()
        .enumerate()
        .all(|(i, gen)| judge_single(r#gen(SEED_BASE ^ i as u64), &program))
}

fn gen_uniform(seed: u64, range: RangeInclusive<Value>) -> Box<[Value]> {
    let rng = Rng::seed_from_u64(seed);
    let uniform = Uniform::new_inclusive(range.start(), range.end());

    rng.sample_iter(uniform).take(CORE_COUNT).collect()
}

fn gen_close(seed: u64) -> Box<[Value]> {
    let mut rng = Rng::seed_from_u64(seed);

    let width: Value = 1 << rng.gen_range(8..=16);
    let start = rng.gen_range(Value::MIN..=Value::MAX - width);
    let end = start + width;

    gen_uniform(rng.r#gen(), start..=end)
}

fn gen_equal(value: Value) -> Box<[Value]> {
    repeat(value).take(CORE_COUNT).collect()
}

fn gen_position(seed: u64, position: usize) -> Box<[Value]> {
    let mut input = gen_uniform(seed, Value::MIN..=Value::MAX);

    let min_pos = input.iter().enumerate().min_by_key(|(_, v)| **v).unwrap().0;
    input.swap(min_pos, position);

    input
}

type Generator = fn(u64) -> Box<[Value]>;

fn input_gens() -> Vec<Generator> {
    let mut inputs: Vec<Generator> = vec![];

    for _ in 0..4 {
        inputs.push(|seed| gen_uniform(seed, 0..=(1 << 16) - 1));
    }

    for _ in 0..16 {
        inputs.push(|seed| gen_uniform(seed, Value::MIN..=Value::MAX));
    }

    inputs.push(|seed| gen_uniform(seed, Value::MIN..=-1));
    inputs.push(|seed| gen_uniform(seed, 1..=Value::MAX));

    for _ in 0..16 {
        inputs.push(gen_close);
    }

    inputs.push(|seed| gen_equal(Rng::seed_from_u64(seed).r#gen()));
    inputs.push(|_| gen_equal(Value::MIN));
    inputs.push(|_| gen_equal(Value::MAX));

    inputs.push(|seed| gen_position(seed, 0));
    inputs.push(|seed| gen_position(seed, 1));
    inputs.push(|seed| gen_position(seed, 2));
    inputs.push(|seed| gen_position(seed, CORE_COUNT - 2));
    inputs.push(|seed| gen_position(seed, CORE_COUNT - 1));

    inputs
}
