use std::{
    cmp::min,
    collections::{HashMap, HashSet},
    iter::repeat,
};

use rand::{seq::SliceRandom, Rng as _, SeedableRng};
use rayon::iter::{IndexedParallelIterator, IntoParallelRefIterator, ParallelIterator};
use sort_machine::{Instruction, Machine, Value};

use crate::{Rng, CORE_COUNT};

const SEED_BASE: u64 = 0xc85e29574c0719d6;

fn correct_answer(input: &[Value]) -> Value {
    let mut counts: HashMap<i64, usize> = HashMap::new();

    for num in input {
        *counts.entry(*num).or_default() += 1;
    }

    counts.into_iter().max_by_key(|(_k, v)| *v).unwrap().0
}

fn judge_single(input: Box<[Value]>, program: &[Instruction]) -> bool {
    let expected = correct_answer(&input);
    let mut machine = Machine::new(CORE_COUNT, Box::new([input]));

    for inst in program {
        machine.exec(*inst);
    }

    machine.output()[0] == expected
}

pub fn judge(program: Vec<Instruction>) -> bool {
    input_gens()
        .par_iter()
        .enumerate()
        .all(|(i, gen)| judge_single(r#gen(SEED_BASE ^ i as u64), &program))
}

fn gen_random(seed: u64, max_largest_count: usize) -> Box<[Value]> {
    let mut rng = Rng::seed_from_u64(seed);

    let mut counts = gen_counts(&mut rng, max_largest_count);
    counts.shuffle(&mut rng);

    input_from_counts(&mut rng, &counts)
}

fn gen_position(seed: u64, pos: isize) -> Box<[Value]> {
    let mut rng = Rng::seed_from_u64(seed);

    let mut counts = gen_counts(&mut rng, CORE_COUNT / 32);
    counts[1..].shuffle(&mut rng);

    if pos >= 0 {
        counts.rotate_right(pos as usize);
    } else {
        counts.rotate_left(-pos as usize);
    }

    input_from_counts(&mut rng, &counts)
}

fn gen_all_equal(seed: u64) -> Box<[Value]> {
    let mut rng = Rng::seed_from_u64(seed);
    let value: Value = rng.gen();

    repeat(value).take(CORE_COUNT).collect()
}

fn gen_counts(rng: &mut Rng, max_largest_count: usize) -> Vec<usize> {
    let largest_count = rng.gen_range(max_largest_count / 2..=max_largest_count);
    let mut remaining_count = CORE_COUNT - largest_count;

    let mut counts = vec![largest_count];

    while remaining_count > 0 {
        let count = rng.gen_range(1..=min(remaining_count, largest_count - 1));
        remaining_count -= count;
        counts.push(count);
    }

    counts
}

fn input_from_counts(rng: &mut Rng, counts: &[usize]) -> Box<[Value]> {
    let mut numbers = Vec::with_capacity(counts.len());
    let mut used_numbers = HashSet::new();

    for _ in 0..counts.len() {
        let number = loop {
            let number: Value = rng.gen();

            if !used_numbers.contains(&number) {
                break number;
            }
        };

        used_numbers.insert(number);
        numbers.push(number);
    }

    numbers.sort();

    counts
        .iter()
        .zip(numbers.iter())
        .flat_map(|(count, number)| repeat(*number).take(*count))
        .collect()
}

type Generator = fn(u64) -> Box<[Value]>;

fn input_gens() -> Vec<Generator> {
    let mut inputs: Vec<Generator> = vec![];

    for _ in 0..16 {
        inputs.push(|seed| gen_random(seed, 32));
    }

    for _ in 0..16 {
        inputs.push(|seed| gen_random(seed, CORE_COUNT - 1));
    }

    for _ in 0..8 {
        inputs.push(|seed| gen_random(seed, 4));
    }

    inputs.push(gen_all_equal);

    inputs.push(|seed| gen_position(seed, 0));
    inputs.push(|seed| gen_position(seed, 1));
    inputs.push(|seed| gen_position(seed, 2));
    inputs.push(|seed| gen_position(seed, -2));
    inputs.push(|seed| gen_position(seed, -1));

    inputs
}
