use std::{
    borrow::Cow,
    env::args_os,
    io::{self, Write},
    path::PathBuf,
    process::exit,
};

use lexopt::{
    Arg::{Long, Short, Value},
    Parser, ValueExt,
};
use owo_colors::OwoColorize;
use thiserror::Error;

use crate::{print_error, BAD_ARGS, IO_ERROR, SUCCESS};

#[derive(Debug, Clone)]
pub struct Args {
    pub program: PathBuf,
    pub input: Option<PathBuf>,
    pub cores: Option<usize>,
    pub output_mode: OutputMode,
    pub resolve_labels: bool,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum OutputMode {
    Output,
    Table,
    Csv,
}

impl Args {
    pub fn parse_or_exit() -> Self {
        fn inner() -> Result<Args, Error> {
            let mut parser = Parser::from_env();

            let mut program: Option<PathBuf> = None;
            let mut input: Option<PathBuf> = None;
            let mut cores: Option<usize> = None;
            let mut output_mode = OutputMode::Output;
            let mut resolve_labels = false;

            while let Some(arg) = parser.next()? {
                match arg {
                    Value(val) if program.is_none() => {
                        program = Some(val.into());
                    }
                    Value(val) if input.is_none() => {
                        input = Some(val.into());
                    }
                    Short('n') | Long("cores") => {
                        cores = Some(parser.value()?.parse()?);
                    }
                    Short('t') | Long("table") => {
                        if output_mode == OutputMode::Output {
                            output_mode = OutputMode::Table;
                        }
                    }
                    Short('c') | Long("csv") => {
                        output_mode = OutputMode::Csv;
                    }
                    Short('r') | Long("resolve") => {
                        resolve_labels = true;
                    }
                    Short('h') | Long("help") => print_help(),
                    _ => return Err(arg.unexpected().into()),
                }
            }

            Ok(Args {
                program: program.ok_or(Error::NoProgram)?,
                input,
                cores,
                output_mode,
                resolve_labels,
            })
        }

        match inner() {
            Ok(args) => args,
            Err(error) => {
                let mut stderr = anstream::stderr().lock();

                let _ = print_error(&mut stderr, error);
                let _ = print_usage(stderr);

                exit(BAD_ARGS.into())
            }
        }
    }
}

fn print_help() {
    if print_usage(anstream::stdout()).is_ok() {
        exit(SUCCESS.into());
    } else {
        exit(IO_ERROR.into());
    }
}

#[derive(Debug, Error)]
enum Error {
    #[error("{}", translate_lexopt_error(.0))]
    Lexing(#[from] lexopt::Error),
    #[error("nebyla uvedena cesta k programu")]
    NoProgram,
}

fn translate_lexopt_error(error: &lexopt::Error) -> String {
    match error {
        lexopt::Error::MissingValue {
            option: Some(option),
        } => format!("přepínač {option} vyžaduje hodnotu"),
        // This shouldn't be possible:
        lexopt::Error::MissingValue { option: None } => "chybí hodnota".to_string(),

        lexopt::Error::UnexpectedOption(option) => format!("neznámý přepínač {option}"),
        lexopt::Error::UnexpectedArgument(argument) => {
            format!("argument „{}“ přebývá", argument.to_string_lossy())
        }
        lexopt::Error::UnexpectedValue { option, value: _ } => {
            format!("přepínač {option} nemůže mít hodnotu")
        }
        lexopt::Error::ParsingFailed { value, error: _ } => {
            format!("hodnota „{value}“ není platná")
        }
        lexopt::Error::NonUnicodeValue(value) => {
            format!(
                "hodnota „{}“ není v kódování UTF-8",
                value.to_string_lossy()
            )
        }

        // This shouldn't be possible:
        lexopt::Error::Custom(error) => error.to_string(),
    }
}

fn print_usage(mut w: impl Write) -> Result<(), io::Error> {
    let program_name = args_os().next();
    let program_name = program_name
        .as_ref()
        .map(|s| s.to_string_lossy())
        .unwrap_or(Cow::Borrowed("sort-machine"));

    writeln!(
        w,
        "{} {} {}",
        "Použití:".bold().underline(),
        program_name.bold(),
        USAGE
    )?;

    writeln!(w)?;
    writeln!(w, "{}", "Argumenty:".bold().underline())?;

    for (name, desc) in ARGS {
        writeln!(w, "  {:9}  {}", name.bold(), desc)?;
    }

    writeln!(w)?;
    writeln!(w, "{}", "Přepínače:".bold().underline())?;

    for (name, desc) in FLAGS {
        writeln!(w, "  {:19}  {}", name.bold(), desc)?;
    }

    Ok(())
}

const USAGE: &str = "[PŘEPÍNAČE] <PROGRAM> [VSTUP]";

#[rustfmt::skip]
const ARGS: &[(&str, &str)] = &[
    ("<PROGRAM>", "Cesta k souboru se zdrojovým kódem programu ke spuštení"),
    ("[VSTUP]", "Cesta k souboru se vstupem"),
];

#[rustfmt::skip]
const FLAGS: &[(&str, &str)] = &[
    ("-n, --cores <JÁDRA>", "Počet jader (potřeba jen pokud je vstup prázdný)"),
    ("-t, --table", "Vypíše tabulku s výsledky všech instrukcí"),
    ("-c, --csv", "Vypíše výsledky všech instrukcí ve formátu CSV"),
    ("-r, --resolve", "Ve výstupu nahradí jména instrukcí jejich čísly"),
    ("-h, --help", "Vypíše tuto nápovědu"),
];
