use std::{
    fmt::Display,
    fs,
    io::{self, stdout, BufWriter, Write},
    path::{Path, PathBuf},
    process::ExitCode,
};

use args::Args;
use csv::print_csv;
use owo_colors::OwoColorize;
use sort_machine::{
    parse, parse_and_format, parse_input, InputError, Instruction, Machine, ParseError, Vector,
};
use table::print_table;
use thiserror::Error;

mod args;
mod csv;
mod table;

const SUCCESS: u8 = 0;
const BAD_PROGRAM: u8 = 1;
const IO_ERROR: u8 = 2;
const BAD_ARGS: u8 = 3;

fn run(args: Args) -> Result<(), CliError> {
    let program = read_file(&args.program)?;

    let input = args.input.as_deref().map(read_file).transpose()?;
    let input = input.as_deref().unwrap_or("");

    let (input, cores) = parse_input(input, args.cores)?;

    let (program, formatted) = if args.resolve_labels {
        let program = parse(&program, input.len()).map_err(CliError::Parse)?;
        let formatted = program.iter().map(Instruction::to_string).collect();
        (program, formatted)
    } else {
        parse_and_format(&program, input.len()).map_err(CliError::Parse)?
    };

    let mut machine = Machine::new(cores, input);

    for inst in program.iter().copied() {
        machine.exec(inst);
    }

    match args.output_mode {
        args::OutputMode::Output => print_output(machine.output())?,
        args::OutputMode::Table => print_table(&machine, &formatted)?,
        args::OutputMode::Csv => print_csv(&machine, &formatted)?,
    }

    Ok(())
}

fn print_output(result: Vector) -> Result<(), CliError> {
    let mut stdout = BufWriter::new(stdout().lock());

    for (i, value) in result.iter().copied().enumerate() {
        if i > 0 {
            stdout.write_all(b" ")?;
        }

        let mut buf = itoa::Buffer::new();
        stdout.write_all(buf.format(value).as_bytes())?;
    }

    stdout.write_all(b"\n")?;
    stdout.flush()?;

    Ok(())
}

fn main() -> ExitCode {
    let args = Args::parse_or_exit();

    match run(args) {
        Ok(()) => ExitCode::from(SUCCESS),
        Err(CliError::Parse(errors)) => {
            let _ = print_parse_errors(anstream::stderr().lock(), &errors);
            ExitCode::from(BAD_PROGRAM)
        }
        Err(error) => {
            let _ = print_error(anstream::stderr().lock(), error);
            ExitCode::from(IO_ERROR)
        }
    }
}

fn read_file(path: &Path) -> Result<String, CliError> {
    fs::read_to_string(path).map_err(|e| CliError::ReadFile(path.to_owned(), e))
}

fn print_error(mut w: impl Write, error: impl Display) -> Result<(), io::Error> {
    writeln!(w, "{header} {error}", header = "Chyba:".red().bold())
}

fn print_parse_errors(mut w: impl Write, errors: &[ParseError]) -> Result<(), io::Error> {
    for error in errors {
        match error {
            ParseError::Instruction { inner, line } => writeln!(
                w,
                "{header} {inner}",
                header = format_args!("Chyba na řádku {line}:").red().bold()
            )?,
            _ => print_error(&mut w, error)?,
        };
    }

    Ok(())
}

#[derive(Debug, Error)]
enum CliError {
    #[error("neplatný program")]
    Parse(Vec<ParseError>),
    #[error(transparent)]
    Input(#[from] InputError),
    #[error("nepodařilo se přečíst \"{0}\": {}", translate_io_error(.1))]
    ReadFile(PathBuf, io::Error),
    #[error("{}", translate_io_error(.0))]
    Io(#[from] io::Error),
}

fn translate_io_error(error: &io::Error) -> String {
    match error.kind() {
        io::ErrorKind::NotFound => "soubor nenalezen".into(),
        io::ErrorKind::PermissionDenied => "přístup zamítnut".into(),
        io::ErrorKind::InvalidData => "soubor není v kódování UTF-8".into(),

        io::ErrorKind::BrokenPipe => "roura přerušena".into(),
        io::ErrorKind::Unsupported => "operace není podporována".into(),
        io::ErrorKind::OutOfMemory => "nedostatek paměti".into(),

        _ => format!("neznámá chyba: {error}"),
    }
}
