use std::io::{self, BufWriter, Write};

use owo_colors::OwoColorize;
use sort_machine::{Machine, Value};
use tabled::{
    builder::Builder,
    settings::{
        object::{Rows, Segment},
        style::{BorderSpanCorrection, HorizontalLine},
        Alignment, Format, Span, Style,
    },
};

pub fn print_table(machine: &Machine, insts: &[String]) -> Result<(), io::Error> {
    let mut builder = Builder::default();

    builder.push_record(
        ["ID".to_owned(), String::default()]
            .into_iter()
            .chain((0..machine.cores()).map(|n| n.to_string())),
    );

    for (i, input) in machine.inputs().iter().enumerate() {
        let i = i + 1;

        builder.push_record(
            [format!("Vstup {i}"), String::default()]
                .into_iter()
                .chain(input.iter().map(Value::to_string)),
        );
    }

    for (i, (result, inst)) in machine.results().iter().zip(insts.iter()).enumerate() {
        let i = i + 1;

        builder.push_record(
            [i.to_string(), inst.to_string()]
                .into_iter()
                .chain(result.iter().map(Value::to_string)),
        );
    }

    let mut table = builder.build();

    let rows = table.count_rows();
    let inputs = machine.inputs().len();

    let style = Style::modern_rounded()
        .horizontals([
            (inputs + 1, HorizontalLine::inherit(Style::modern_rounded())),
            (rows - 1, HorizontalLine::inherit(Style::modern_rounded())),
        ])
        .remove_horizontal();

    table.with(style);

    table.modify(Segment::new(..inputs + 1, 0..=0), Span::column(2));
    table.with(BorderSpanCorrection);

    table.modify(Segment::new(.., 2..), Alignment::right());
    table.modify(Segment::new(inputs + 1.., 0..=0), Alignment::right());

    table.modify(
        Rows::new(inputs + 1..rows - 1),
        Format::positioned(|c, (x, _y)| {
            if (x - inputs) % 2 == 0 {
                c.dimmed().to_string()
            } else {
                c.to_string()
            }
        }),
    );

    table.modify(
        Rows::new(..inputs + 1),
        Format::content(|c| c.bold().to_string()),
    );
    table.modify(Rows::last(), Format::content(|c| c.bold().to_string()));

    let mut stdout = BufWriter::new(anstream::stdout().lock());
    writeln!(stdout, "{table}")?;

    stdout.flush()?;

    Ok(())
}
