use std::{future::pending, net::SocketAddr, path::PathBuf, sync::Arc};

use anyhow::Result;
use axum::{
    extract::State,
    http::{
        header::{CACHE_CONTROL, REFERRER_POLICY, X_CONTENT_TYPE_OPTIONS},
        HeaderValue,
    },
    response::Redirect,
    routing::{get, post},
    Form, Router,
};
use clap::Parser;
use limits::INITIAL_PERMITS;
use maud::Markup;
use simulation::{
    parse_submission, simulate_with_output, simulate_with_tracing, Action, SimulationResult,
    Submission,
};
use tokio::{net::TcpListener, signal::ctrl_c, sync::Semaphore};
use tower::ServiceBuilder;
use tower_http::{
    compression::CompressionLayer, services::ServeDir, set_header::SetResponseHeaderLayer,
    trace::TraceLayer,
};
use tracing::{error, info, level_filters::LevelFilter};
use tracing_subscriber::{fmt, layer::SubscriberExt, registry, util::SubscriberInitExt, Layer};
use view::{download, editor};

mod limits;
mod simulation;
mod view;

#[derive(Debug, Clone)]
struct SimulatorState {
    pub semaphore: Arc<Semaphore>,
}

async fn get_editor() -> Markup {
    editor::page(None, None)
}

async fn post_editor(state: State<SimulatorState>, form: Form<Submission>) -> Markup {
    let result = match parse_submission(&form) {
        Ok(parsed) => {
            let _permit = state
                .semaphore
                .acquire_many(parsed.permits_needed())
                .await
                .unwrap();

            info!(
                action=?form.action,
                instructions=parsed.program.len(),
                cores=parsed.machine.cores(),
                "Running simulation"
            );

            match form.action {
                Action::Run => SimulationResult::Output(simulate_with_output(parsed)),
                Action::Trace => SimulationResult::Tracing(simulate_with_tracing(parsed)),
            }
        }
        Err(errors) => SimulationResult::Errors(errors),
    };

    editor::page(Some(result), Some(&form))
}

async fn download() -> Markup {
    download::page()
}

#[derive(Debug, Parser)]
/// A web-based simulator for a custom computational model
struct Args {
    /// The port to listen on
    #[arg(long, default_value_t = 8000, env)]
    port: u16,
    /// The path to static assets
    #[arg(long, env)]
    r#static: PathBuf,
    /// Allow browsers and proxies to cache static files
    #[arg(long, env)]
    cache_static: bool,
}

#[tokio::main]
async fn run_server(args: Args) -> Result<()> {
    let state = SimulatorState {
        semaphore: Arc::new(Semaphore::new(INITIAL_PERMITS)),
    };

    let app = Router::new()
        .route("/", get(get_editor))
        .route("/", post(post_editor))
        .route("/download", get(download))
        .route(
            "/download/",
            get(|| async { Redirect::permanent("/download") }),
        );

    let app = app.fallback_service(
        ServiceBuilder::new()
            .layer(SetResponseHeaderLayer::if_not_present(
                CACHE_CONTROL,
                if args.cache_static {
                    HeaderValue::from_static("public, max-age=600")
                } else {
                    HeaderValue::from_static("no-cache")
                },
            ))
            .service(ServeDir::new(args.r#static)),
    );

    let app = app
        .layer(SetResponseHeaderLayer::if_not_present(
            X_CONTENT_TYPE_OPTIONS,
            HeaderValue::from_static("nosniff"),
        ))
        .layer(SetResponseHeaderLayer::if_not_present(
            REFERRER_POLICY,
            HeaderValue::from_static("strict-origin-when-cross-origin"),
        ))
        .layer(CompressionLayer::new().gzip(true).deflate(true).br(true))
        .layer(TraceLayer::new_for_http())
        .with_state(state);

    let address = SocketAddr::from(([0, 0, 0, 0, 0, 0, 0, 0], args.port));
    let listener = TcpListener::bind(address).await?;

    info!(%address, "Server started");

    axum::serve(listener, app.into_make_service())
        .with_graceful_shutdown(wait_for_termination())
        .await?;

    Ok(())
}

fn main() {
    let args = Args::parse();

    let stdout = fmt::layer().with_filter(LevelFilter::INFO);
    registry().with(stdout).init();

    match run_server(args) {
        Ok(()) => {}
        Err(error) => error!("{error}"),
    }
}

async fn wait_for_termination() {
    let ctrl_c = async {
        if let Err(error) = ctrl_c().await {
            error!("Failed to install signal handler: {error}");
            pending::<()>().await;
        }
    };

    #[cfg(unix)]
    {
        use tokio::{
            select,
            signal::unix::{signal, SignalKind},
        };

        let sigterm = async {
            match signal(SignalKind::terminate()) {
                Ok(mut stream) => {
                    stream.recv().await;
                }
                Err(error) => {
                    error!("Failed to install signal handler: {error}");
                    pending::<()>().await;
                }
            }
        };

        select! {
            _ = ctrl_c => {}
            _ = sigterm => {}
        }
    }

    #[cfg(not(unix))]
    ctrl_c.await;

    info!("Stopping server, waiting for requests to finish")
}
