use std::{
    fmt::{self, Display, Formatter},
    iter::repeat,
};

use serde::{Deserialize, Deserializer};
use sort_machine::{
    parse_and_format, parse_input, InputError, Instruction, Machine, ParseError, Value, Vector,
};
use thiserror::Error;

#[derive(Debug, Deserialize, Clone)]
pub struct Submission {
    pub program: String,
    pub input: String,
    #[serde(deserialize_with = "parse_optional_int")]
    pub cores: Option<usize>,
    pub action: Action,
}

#[derive(Debug, Deserialize, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
#[serde(rename_all = "kebab-case")]
pub enum Action {
    Trace,
    Run,
}

#[derive(Debug)]
pub enum SimulationResult {
    Output(Box<[Value]>),
    Tracing(TracingOutput),
    Errors(Vec<SimulationError>),
}

#[derive(Debug, Clone)]
pub struct ParseOutput {
    pub machine: Machine,
    pub program: Vec<Instruction>,
    pub formatted: Vec<String>,
}

#[derive(Debug, Clone)]
pub struct TracingOutput {
    pub cores: usize,
    pub inputs: Box<[Vector]>,
    pub insts: Box<[TracingRow]>,
}

#[derive(Debug, Clone)]
pub struct TracingRow {
    pub instruction: String,
    pub values: Vector,
    pub order: Option<Box<[usize]>>,
}

pub fn parse_submission(submission: &Submission) -> Result<ParseOutput, Vec<SimulationError>> {
    let (inputs, cores) =
        parse_input(&submission.input, submission.cores).map_err(|e| vec![e.into()])?;

    let (program, formatted) = parse_and_format(&submission.program, inputs.len())
        .map_err(|e| e.into_iter().map(SimulationError::from).collect::<Vec<_>>())?;

    let output = ParseOutput {
        machine: Machine::new(cores, inputs),
        program,
        formatted,
    };

    output
        .check_limits(submission.action)
        .map_err(|e| vec![e])?;

    Ok(output)
}

pub fn simulate_with_output(
    ParseOutput {
        mut machine,
        program,
        ..
    }: ParseOutput,
) -> Box<[Value]> {
    for inst in program {
        machine.exec(inst);
    }

    machine.output().iter().copied().collect()
}

pub fn simulate_with_tracing(
    ParseOutput {
        mut machine,
        program,
        formatted,
    }: ParseOutput,
) -> TracingOutput {
    let mut rows = Vec::with_capacity(program.len());

    for (inst, formatted) in program.into_iter().zip(formatted.into_iter()) {
        machine.exec(inst);

        let order = if matches!(inst, Instruction::Sort(_)) {
            let mut order: Box<[usize]> = repeat(0).take(machine.cores()).collect();

            for (pos, id) in machine.cores_by_pos().iter().copied().enumerate() {
                order[id] = pos;
            }

            Some(order)
        } else {
            None
        };

        rows.push(TracingRow {
            instruction: formatted,
            values: machine.output(),
            order,
        });
    }

    TracingOutput {
        cores: machine.cores(),
        inputs: machine.inputs().into(),
        insts: rows.into_boxed_slice(),
    }
}

fn parse_optional_int<'de, D>(deserializer: D) -> Result<Option<usize>, D::Error>
where
    D: Deserializer<'de>,
{
    Ok(<&str>::deserialize(deserializer)?.parse().ok())
}

impl Display for Action {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            Action::Trace => write!(f, "trasovat"),
            Action::Run => write!(f, "spustit"),
        }
    }
}

#[derive(Debug, Error)]
pub enum SimulationError {
    #[error(transparent)]
    Parse(#[from] ParseError),
    #[error(transparent)]
    Input(#[from] InputError),
    #[error("není možné {action} program, který by spotřeboval přes {mebibytes} MiB paměti")]
    MemoryLimit { action: Action, mebibytes: usize },
    #[error("není možné spustit program na více než {cores} jádrech")]
    CoreLimit { cores: usize },
}
