use maud::{html, Markup};
use sort_machine::{ParseError, Value};

use crate::{
    simulation::{SimulationError, SimulationResult, TracingOutput, TracingRow},
    Submission,
};

use super::document;

fn editor(result: Option<SimulationResult>, prev_submission: Option<&Submission>) -> Markup {
    let cores = prev_submission.and_then(|s| s.cores);

    html!(
        main."with-output"[result.is_some()] {
            form method="POST" {
                .inputs."save-size" #"inputs-area" {
                    label.panel.editor.program."save-size" #"program-area" for="program" {
                        .label { "Program" }

                        textarea."save-scroll" #program name="program" {
                            @if let Some(sub) = &prev_submission {
                                (sub.program)
                            }
                        }
                    }

                    label.panel.editor.input for="input" {
                        .label { "Vstup" }

                        textarea.save-scroll #input name="input" {
                            @if let Some(sub) = &prev_submission {
                                (sub.input)
                            }
                        }
                    }
                }

                @if let Some(result) = result {
                    @match result {
                        SimulationResult::Output(o) => (output(&o)),
                        SimulationResult::Tracing(t) => (table(&t)),
                        SimulationResult::Errors(e) => (errors(&e)),
                    }
                }

                .controls {
                    button name="action" value="run" {
                        "Spustit"
                    }
                    button name="action" value="trace" {
                        "Trasovat"
                    }

                    label.cores {
                        "Počet jader: "
                        input type="number" name="cores" value=[cores];
                    }
                }
            }
        }
    )
}

fn output(output: &[Value]) -> Markup {
    let mut buf = itoa::Buffer::new();

    html!(
        label.panel.editor.output for="output" {
            .label { "Výstup" }

            textarea."save-scroll" #output readonly {
                @for (i, value) in output.iter().copied().enumerate() {
                    @if i > 0 {
                        " ";
                    }

                    (buf.format(value))
                }
            }
        }
    )
}

fn errors(errors: &[SimulationError]) -> Markup {
    html!(
        .panel.output {
            .label { "Výstup" }

            .errors {
                @for error in errors {
                    @match error {
                        SimulationError::Parse(ParseError::Instruction { inner, line }) => p.error {
                            b { "Chyba na řádku " (line) ":" }
                            " " (inner)
                        },
                        error => p.error {
                            b { "Chyba:" }
                            " " (error)
                        },
                    }
                }
            }
        }
    )
}

fn table(output: &TracingOutput) -> Markup {
    html!(
        .panel.output {
            .label { "Výstup" }

            .scroll."save-scroll" #tracing {
                table {
                    thead {
                        tr.sortable {
                            th { "ID" }

                            @for i in 0..output.cores {
                                td "data-pos"=(i) { (i) }
                            }
                        }

                        @for (i, input) in output.inputs.iter().enumerate() {
                            tr {
                                th { "Vstup " (i + 1) }

                                @for value in input.iter() {
                                    td { (value) }
                                }
                            }
                        }
                    }

                    tbody {
                        @for (i, TracingRow { instruction, values, order }) in output.insts.iter().enumerate() {
                            tr.sortable[order.is_some()] {
                                th {
                                    (i + 1)". "
                                    span.instruction { (instruction) }
                                }

                                @if let Some(order) = order {
                                    @for (value, pos) in values.iter().zip(order) {
                                        td "data-pos"=(pos) { (value) }
                                    }
                                } @else {
                                    @for value in values.iter() {
                                        td { (value) }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    )
}

pub fn page(result: Option<SimulationResult>, prev_submission: Option<&Submission>) -> Markup {
    document(
        editor(result, prev_submission),
        html!(
            link rel="stylesheet" href="assets/editor.css";
            script defer src="assets/editor.js" {}
        ),
        "Třídicí procesor",
        "Simulátor pro výpočetní model z MO-P 74",
    )
}
