use thiserror::Error;

use crate::Value;

type Inputs = Box<[Box<[Value]>]>;

pub fn parse_input(
    table: &str,
    core_count_hint: Option<usize>,
) -> Result<(Inputs, usize), InputError> {
    let input = parse_table(table)?;

    if let Some(first_input) = input.first() {
        let cores = first_input.len();

        if input.iter().any(|input| input.len() != cores) {
            return Err(InputError::LineLengthMismatch);
        }

        if let Some(hint) = core_count_hint {
            if hint != cores {
                return Err(InputError::CoreCountMismatch {
                    columns: cores,
                    cores: hint,
                });
            }
        }

        Ok((input, cores))
    } else if let Some(cores) = core_count_hint {
        Ok((Box::new([]), cores))
    } else {
        Err(InputError::NoInputOrCoreCount)
    }
}

fn parse_table(table: &str) -> Result<Inputs, InputError> {
    let mut inputs: Vec<Box<[i64]>> = vec![];

    for line in table.lines() {
        let tokens = line.split_whitespace();
        let mut input = Vec::with_capacity(inputs.first().map(|l| l.len()).unwrap_or(0));

        for token in tokens {
            let number = token
                .parse()
                .map_err(|_| InputError::InvalidInt(token.into()))?;

            input.push(number);
        }

        if !input.is_empty() {
            inputs.push(input.into_boxed_slice());
        }
    }

    Ok(inputs.into_boxed_slice())
}

#[derive(Debug, Error)]
pub enum InputError {
    #[error("některé řádky mají různý počet hodnot")]
    LineLengthMismatch,
    #[error("vstup má {columns} sloupců, ale procesor má {cores} jader")]
    CoreCountMismatch { cores: usize, columns: usize },
    #[error("není možné odvodit počet jader ze vstupu, protože je prázný")]
    NoInputOrCoreCount,

    #[error("„{0}“ není platné číslo")]
    InvalidInt(Box<str>),
}
