use std::fmt::{self, Display};

use crate::Value;

#[derive(Debug, Clone, Copy)]
pub enum Instruction {
    Const(Value),
    Id,
    Input(usize),
    Copy(usize),
    BinaryOp {
        op: BinaryOp,
        lhs: usize,
        rhs: usize,
    },
    If {
        cond: usize,
        r#true: usize,
        r#false: usize,
    },
    ReadSide {
        side: Side,
        value: usize,
    },
    Sort(usize),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum BinaryOp {
    Add,
    Sub,
    Mul,
    Div,
    Rem,
    Or,
    And,
    Xor,
    Eq,
    NotEq,
    Less,
    Greater,
    LessEq,
    GreaterEq,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum Side {
    Left,
    Right,
}

impl Display for Instruction {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Instruction::Const(c) => write!(f, "const {c}"),
            Instruction::Id => write!(f, "id"),
            Instruction::Input(i) => write!(f, "input {}", i + 1),
            Instruction::Copy(i) => write!(f, "copy {}", i + 1),
            Instruction::BinaryOp { op, lhs, rhs } => write!(f, "{op} {} {}", lhs + 1, rhs + 1),
            Instruction::If {
                cond,
                r#true,
                r#false,
            } => write!(f, "if {} {} {}", cond + 1, r#true + 1, r#false + 1),
            Instruction::ReadSide { side, value } => write!(f, "{side} {}", value + 1),
            Instruction::Sort(k) => write!(f, "sort {}", k + 1),
        }
    }
}

impl Display for BinaryOp {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            BinaryOp::Add => write!(f, "+"),
            BinaryOp::Sub => write!(f, "-"),
            BinaryOp::Mul => write!(f, "*"),
            BinaryOp::Div => write!(f, "/"),
            BinaryOp::Rem => write!(f, "%"),
            BinaryOp::Or => write!(f, "|"),
            BinaryOp::And => write!(f, "&"),
            BinaryOp::Xor => write!(f, "^"),
            BinaryOp::Eq => write!(f, "="),
            BinaryOp::NotEq => write!(f, "!="),
            BinaryOp::Less => write!(f, "<"),
            BinaryOp::Greater => write!(f, ">"),
            BinaryOp::LessEq => write!(f, "<="),
            BinaryOp::GreaterEq => write!(f, ">="),
        }
    }
}

impl Display for Side {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Side::Left => write!(f, "left"),
            Side::Right => write!(f, "right"),
        }
    }
}
