use crate::{
    instruction::{BinaryOp, Instruction, Side},
    Value, Vector,
};

#[derive(Debug, Clone)]
pub struct Machine {
    cores: usize,

    inputs: Box<[Vector]>,
    results: Vec<Vector>,

    cores_by_pos: Box<[usize]>,
}

impl Machine {
    pub fn new(cores: usize, inputs: Box<[Box<[Value]>]>) -> Self {
        for input in &inputs {
            assert_eq!(input.len(), cores);
        }

        Machine {
            cores,
            inputs: inputs.into_vec().into_iter().map(Vector::from).collect(),
            results: Vec::new(),
            cores_by_pos: (0..cores).collect(),
        }
    }

    pub fn exec(&mut self, inst: Instruction) {
        let result = match inst {
            Instruction::Const(value) => self.exec_const(value),
            Instruction::Id => self.exec_id(),
            Instruction::Input(idx) => self.exec_input(idx),
            Instruction::Copy(idx) => self.exec_copy(idx),
            Instruction::BinaryOp { op, lhs, rhs } => self.exec_binary_op(op, lhs, rhs),
            Instruction::If {
                cond,
                r#true,
                r#false,
            } => self.exec_if(cond, r#true, r#false),
            Instruction::ReadSide { side, value } => self.exec_read_side(side, value),
            Instruction::Sort(key) => self.exec_sort(key),
        };

        self.results.push(result);
    }

    fn exec_const(&self, value: Value) -> Vector {
        Vector::Const {
            value,
            len: self.cores(),
        }
    }

    fn exec_id(&self) -> Vector {
        let cores = self.cores.try_into().unwrap();
        (0..cores).collect()
    }

    fn exec_input(&self, idx: usize) -> Vector {
        self.inputs[idx].clone()
    }

    fn exec_copy(&self, idx: usize) -> Vector {
        self.results[idx].clone()
    }

    fn exec_binary_op(&self, op: BinaryOp, lhs: usize, rhs: usize) -> Vector {
        fn inner(lhs: &Vector, rhs: &Vector, f: impl Fn(Value, Value) -> Value) -> Vector {
            match (lhs, rhs) {
                (Vector::Array(lhs), Vector::Array(rhs)) => {
                    lhs.iter().zip(rhs.iter()).map(|(a, b)| f(*a, *b)).collect()
                }
                (Vector::Array(lhs), Vector::Const { value: rhs, .. }) => {
                    lhs.iter().map(|a| f(*a, *rhs)).collect()
                }
                (Vector::Const { value: lhs, .. }, Vector::Array(rhs)) => {
                    rhs.iter().map(|b| f(*lhs, *b)).collect()
                }
                (Vector::Const { value: lhs, .. }, Vector::Const { value: rhs, len }) => {
                    Vector::Const {
                        value: f(*lhs, *rhs),
                        len: *len,
                    }
                }
            }
        }

        let lhs = &self.results[lhs];
        let rhs = &self.results[rhs];

        match op {
            BinaryOp::Add => inner(lhs, rhs, |a, b| a.wrapping_add(b)),
            BinaryOp::Sub => inner(lhs, rhs, |a, b| a.wrapping_sub(b)),
            BinaryOp::Mul => inner(lhs, rhs, |a, b| a.wrapping_mul(b)),
            BinaryOp::Div => inner(lhs, rhs, |a, b| if b != 0 { a.wrapping_div(b) } else { 0 }),
            BinaryOp::Rem => inner(lhs, rhs, |a, b| if b != 0 { a.wrapping_rem(b) } else { 0 }),

            BinaryOp::Or => inner(lhs, rhs, |a, b| a | b),
            BinaryOp::And => inner(lhs, rhs, |a, b| a & b),
            BinaryOp::Xor => inner(lhs, rhs, |a, b| a ^ b),

            BinaryOp::Eq => inner(lhs, rhs, |a, b| if a == b { 1 } else { 0 }),
            BinaryOp::NotEq => inner(lhs, rhs, |a, b| if a != b { 1 } else { 0 }),
            BinaryOp::Less => inner(lhs, rhs, |a, b| if a < b { 1 } else { 0 }),
            BinaryOp::Greater => inner(lhs, rhs, |a, b| if a > b { 1 } else { 0 }),
            BinaryOp::LessEq => inner(lhs, rhs, |a, b| if a <= b { 1 } else { 0 }),
            BinaryOp::GreaterEq => inner(lhs, rhs, |a, b| if a >= b { 1 } else { 0 }),
        }
    }

    fn exec_if(&self, cond: usize, r#true: usize, r#false: usize) -> Vector {
        let cond = &self.results[cond];
        let r#true = &self.results[r#true];
        let r#false = &self.results[r#false];

        cond.iter()
            .zip(r#true.iter().zip(r#false.iter()))
            .map(|(cond, (r#true, r#false))| if *cond != 0 { *r#true } else { *r#false })
            .collect()
    }

    fn exec_read_side(&self, side: Side, value: usize) -> Vector {
        let value = &self.results[value];

        let value = match value {
            Vector::Array(a) => a,
            Vector::Const { .. } => return value.clone(),
        };

        let mut result = vec![0; self.cores].into_boxed_slice();

        if self.cores == 0 {
            return result.into();
        }

        match side {
            Side::Left => {
                for pos in 1..self.cores {
                    result[self.cores_by_pos[pos]] = value[self.cores_by_pos[pos - 1]];
                }

                result[self.cores_by_pos[0]] = value[self.cores_by_pos[self.cores - 1]];
            }
            Side::Right => {
                for pos in 0..self.cores - 1 {
                    result[self.cores_by_pos[pos]] = value[self.cores_by_pos[pos + 1]];
                }

                result[self.cores_by_pos[self.cores - 1]] = value[self.cores_by_pos[0]];
            }
        }

        result.into()
    }

    fn exec_sort(&mut self, key: usize) -> Vector {
        let key = &self.results[key];

        match key {
            Vector::Array(key) => self.cores_by_pos.sort_by_key(|id| key[*id]),
            Vector::Const { .. } => {}
        }

        key.clone()
    }

    pub fn cores(&self) -> usize {
        self.cores
    }

    pub fn inputs(&self) -> &[Vector] {
        &self.inputs
    }

    pub fn output(&self) -> Vector {
        self.results.last().unwrap().clone()
    }

    pub fn results(&self) -> &[Vector] {
        &self.results
    }

    pub fn cores_by_pos(&self) -> &[usize] {
        &self.cores_by_pos
    }
}
