use std::{collections::HashMap, mem::take, str::FromStr};

use thiserror::Error;

use crate::instruction::{BinaryOp, Instruction, Side};

fn parse_internal(
    program: &str,
    input_count: usize,
    mut formatted: Option<&mut Vec<String>>,
) -> Result<Vec<Instruction>, Vec<ParseError>> {
    let mut insts = Vec::new();
    let mut inst_count = 0usize;

    let mut labels = HashMap::new();
    let mut next_inst_labels = Vec::new();

    let mut errors = Vec::new();

    for (line_num, mut line) in program.lines().enumerate() {
        let line_num = line_num + 1;
        let id = insts.len();

        if let Some((inst, _comment)) = line.split_once('#') {
            line = inst;
        }

        while let Some((label, rest)) = line.split_once(':') {
            let label = label.trim();

            if !is_valid_ident(label) {
                errors.push(ParseError::Instruction {
                    inner: ParseInstructionError::InvalidIdent(label.into()),
                    line: line_num,
                });
            }

            next_inst_labels.push(label);

            line = rest;
        }

        match parse_inst(line, &labels, inst_count, input_count) {
            Ok(None) => continue,
            Ok(Some(inst)) => insts.push(inst),
            Err(inner) => errors.push(ParseError::Instruction {
                inner,
                line: line_num,
            }),
        }

        inst_count += 1;

        if let Some(ref mut formatted) = formatted {
            formatted.push(format_inst(
                next_inst_labels.iter().copied(),
                line.split_whitespace(),
            ));
        }

        for label in take(&mut next_inst_labels) {
            labels.insert(label, id);
        }
    }

    if insts.is_empty() && errors.is_empty() {
        errors.push(ParseError::EmptyProgram);
    }

    if !errors.is_empty() {
        Err(errors)
    } else {
        Ok(insts)
    }
}

pub fn parse(program: &str, input_count: usize) -> Result<Vec<Instruction>, Vec<ParseError>> {
    parse_internal(program, input_count, None)
}

pub fn parse_and_format(
    program: &str,
    input_count: usize,
) -> Result<(Vec<Instruction>, Vec<String>), Vec<ParseError>> {
    let mut formatted = Vec::new();
    let result = parse_internal(program, input_count, Some(&mut formatted));
    result.map(|program| (program, formatted))
}

fn parse_inst(
    line: &str,
    labels: &HashMap<&str, usize>,
    insts: usize,
    inputs: usize,
) -> Result<Option<Instruction>, ParseInstructionError> {
    let mut tokens = line.split_whitespace();

    let Some(mnemonic) = tokens.next() else {
        return Ok(None);
    };

    let arg_count = tokens.clone().count();

    let check_arity = |arity| {
        if arg_count != arity {
            Err(ParseInstructionError::WrongArity {
                mnemonic: mnemonic.into(),
                expected: arity,
                value: arg_count,
            })
        } else {
            Ok(())
        }
    };

    let mut parse_binary_op = |op| {
        check_arity(2)?;

        Ok(Instruction::BinaryOp {
            op,
            lhs: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
            rhs: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
        })
    };

    let inst = match mnemonic {
        "const" => {
            check_arity(1)?;
            Instruction::Const(parse_number(tokens.next().unwrap())?)
        }
        "id" => {
            check_arity(0)?;
            Instruction::Id
        }
        "input" => {
            check_arity(1)?;
            let idx: usize = parse_number(tokens.next().unwrap())?;

            if idx < 1 || idx > inputs {
                return Err(ParseInstructionError::UnknownInput {
                    max: inputs,
                    value: idx,
                });
            }

            Instruction::Input(idx - 1)
        }

        "copy" => {
            check_arity(1)?;
            Instruction::Copy(parse_inst_ref(tokens.next().unwrap(), labels, insts)?)
        }

        "sort" => {
            check_arity(1)?;
            Instruction::Sort(parse_inst_ref(tokens.next().unwrap(), labels, insts)?)
        }
        "left" => {
            check_arity(1)?;
            Instruction::ReadSide {
                side: Side::Left,
                value: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
            }
        }
        "right" => {
            check_arity(1)?;
            Instruction::ReadSide {
                side: Side::Right,
                value: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
            }
        }

        "+" => parse_binary_op(BinaryOp::Add)?,
        "-" => parse_binary_op(BinaryOp::Sub)?,
        "*" => parse_binary_op(BinaryOp::Mul)?,
        "/" => parse_binary_op(BinaryOp::Div)?,
        "%" => parse_binary_op(BinaryOp::Rem)?,
        "|" => parse_binary_op(BinaryOp::Or)?,
        "&" => parse_binary_op(BinaryOp::And)?,
        "^" => parse_binary_op(BinaryOp::Xor)?,
        "=" => parse_binary_op(BinaryOp::Eq)?,
        "!=" => parse_binary_op(BinaryOp::NotEq)?,
        "<" => parse_binary_op(BinaryOp::Less)?,
        ">" => parse_binary_op(BinaryOp::Greater)?,
        "<=" => parse_binary_op(BinaryOp::LessEq)?,
        ">=" => parse_binary_op(BinaryOp::GreaterEq)?,

        "if" => {
            check_arity(3)?;
            Instruction::If {
                cond: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
                r#true: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
                r#false: parse_inst_ref(tokens.next().unwrap(), labels, insts)?,
            }
        }

        _ => return Err(ParseInstructionError::UnknownMnemonic(mnemonic.into())),
    };

    Ok(Some(inst))
}

fn parse_number<T>(token: &str) -> Result<T, ParseInstructionError>
where
    T: FromStr,
{
    token
        .parse()
        .map_err(|_| ParseInstructionError::InvalidNumber(token.into()))
}

fn parse_inst_ref(
    token: &str,
    labels: &HashMap<&str, usize>,
    insts: usize,
) -> Result<usize, ParseInstructionError> {
    if let Ok(num) = token.parse::<usize>() {
        if num >= 1 && num <= insts {
            Ok(num - 1)
        } else {
            Err(ParseInstructionError::InstructionRefOutOfRange {
                max: insts,
                value: num,
            })
        }
    } else if let Some(inst) = labels.get(token) {
        Ok(*inst)
    } else {
        Err(ParseInstructionError::UnknownLabel(token.into()))
    }
}

fn is_valid_ident(ident: &str) -> bool {
    ident.chars().all(|c| c.is_alphanumeric() || c == '_') && !ident.chars().all(char::is_numeric)
}

fn format_inst<'a>(
    labels: impl Iterator<Item = &'a str>,
    tokens: impl Iterator<Item = &'a str>,
) -> String {
    let mut string = String::new();

    for label in labels {
        if !string.is_empty() {
            string.push(' ');
        }

        string.push_str(label);
        string.push(':');
    }

    for token in tokens {
        if !string.is_empty() {
            string.push(' ');
        }

        string.push_str(token);
    }

    string
}

#[derive(Debug, Error)]
pub enum ParseInstructionError {
    #[error("„{0}“ není platný název návěšti")]
    InvalidIdent(Box<str>),
    #[error("instrukce s názvem „{0}“ neexistuje")]
    UnknownMnemonic(Box<str>),
    #[error("instrukce {mnemonic} potřebuje {expected} argumentů, ale dostala {value}")]
    WrongArity {
        mnemonic: Box<str>,
        expected: usize,
        value: usize,
    },
    #[error("výsledek instrukce {value} neexistuje, instrukcí je {max}")]
    InstructionRefOutOfRange { max: usize, value: usize },
    #[error("návěšť „{0}“ nebyla definována")]
    UnknownLabel(Box<str>),
    #[error("„{0}“ není platné číslo")]
    InvalidNumber(Box<str>),
    #[error("vstup číslo {value} neexistuje, vstupů je {max}")]
    UnknownInput { max: usize, value: usize },
}

#[derive(Debug, Error)]
pub enum ParseError {
    #[error("na řádku {line}: {inner}")]
    Instruction {
        inner: ParseInstructionError,
        line: usize,
    },
    #[error("program je prázdný")]
    EmptyProgram,
}
