use crate::{BinaryOp, Instruction, Machine, Side, Value, Vector};

#[test]
fn execute_const() {
    let mut machine = Machine::new(2, input(&[]));

    machine.exec(Instruction::Const(1234));

    assert_eq!(machine.output(), output(&[1234, 1234]));
}

#[test]
fn execute_id() {
    let mut machine = Machine::new(4, input(&[]));

    machine.exec(Instruction::Id);

    assert_eq!(machine.output(), output(&[0, 1, 2, 3]));
}

#[test]
fn execute_input() {
    let mut machine = Machine::new(3, input(&[&[1, 2, 3], &[4, 5, 6]]));

    machine.exec(Instruction::Input(1));

    assert_eq!(machine.output(), output(&[4, 5, 6]));
}

#[test]
fn execute_copy() {
    let mut machine = Machine::new(3, input(&[&[1, 2, 3]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Copy(0));

    assert_eq!(machine.output(), output(&[1, 2, 3]));
}

#[test]
fn execute_sort() {
    let mut machine = Machine::new(4, input(&[&[5, 20, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Sort(0));

    assert_eq!(machine.cores_by_pos(), &[3, 0, 2, 1]);
}

#[test]
fn execute_sort_stable() {
    let mut machine = Machine::new(5, input(&[&[0, 10, 20, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Sort(0));

    assert_eq!(machine.cores_by_pos(), &[0, 4, 1, 3, 2]);
}

#[test]
fn execute_left_many() {
    let mut machine = Machine::new(4, input(&[&[1, 2, 3, 4]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::ReadSide {
        side: Side::Left,
        value: 0,
    });

    assert_eq!(machine.output(), output(&[4, 1, 2, 3]));
}

#[test]
fn execute_left_one() {
    let mut machine = Machine::new(1, input(&[&[10]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::ReadSide {
        side: Side::Left,
        value: 0,
    });

    assert_eq!(machine.output(), output(&[10]));
}

#[test]
fn execute_left_sorted() {
    let mut machine = Machine::new(4, input(&[&[4, 3, 1, 2], &[1, 2, 3, 4]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Sort(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::ReadSide {
        side: Side::Left,
        value: 2,
    });

    assert_eq!(machine.output(), output(&[2, 4, 1, 3]));
}

#[test]
fn execute_right_many() {
    let mut machine = Machine::new(4, input(&[&[1, 2, 3, 4]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::ReadSide {
        side: Side::Right,
        value: 0,
    });

    assert_eq!(machine.output(), output(&[2, 3, 4, 1]));
}

#[test]
fn execute_right_one() {
    let mut machine = Machine::new(1, input(&[&[10]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::ReadSide {
        side: Side::Right,
        value: 0,
    });

    assert_eq!(machine.output(), output(&[10]));
}

#[test]
fn execute_right_sorted() {
    let mut machine = Machine::new(4, input(&[&[4, 3, 1, 2], &[1, 2, 3, 4]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Sort(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::ReadSide {
        side: Side::Right,
        value: 2,
    });

    assert_eq!(machine.output(), output(&[3, 1, 4, 2]));
}

#[test]
fn execute_add() {
    let mut machine = Machine::new(2, input(&[&[1, -10], &[2, 20]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Add,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[3, 10]));
}

#[test]
fn execute_sub() {
    let mut machine = Machine::new(2, input(&[&[1, -10], &[2, 20]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Sub,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[-1, -30]));
}

#[test]
fn execute_mul() {
    let mut machine = Machine::new(2, input(&[&[1, -10], &[2, 20]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Mul,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[2, -200]));
}

#[test]
fn execute_div() {
    let mut machine = Machine::new(3, input(&[&[20, 19, -19], &[5, -5, 5]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Div,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[4, -3, -3]));
}

#[test]
fn execute_div_zero() {
    let mut machine = Machine::new(2, input(&[&[1, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Const(0));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Div,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 0]));
}

#[test]
fn execute_rem() {
    let mut machine = Machine::new(3, input(&[&[20, 19, -19], &[5, -5, 5]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Rem,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 4, -4]));
}

#[test]
fn execute_rem_zero() {
    let mut machine = Machine::new(2, input(&[&[1, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Const(0));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Rem,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 0]));
}

#[test]
fn execute_and() {
    let mut machine = Machine::new(2, input(&[&[9, 5], &[3, -1]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::And,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[1, 5]));
}

#[test]
fn execute_or() {
    let mut machine = Machine::new(2, input(&[&[9, 5], &[3, -1]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Or,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[11, -1]));
}

#[test]
fn execute_xor() {
    let mut machine = Machine::new(2, input(&[&[9, 5], &[3, -1]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Xor,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[10, !5]));
}

#[test]
fn execute_eq() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Eq,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[1, 0, 0]));
}

#[test]
fn execute_not_eq() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::NotEq,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 1, 1]));
}

#[test]
fn execute_less() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Less,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 1, 0]));
}

#[test]
fn execute_less_eq() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::LessEq,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[1, 1, 0]));
}

#[test]
fn execute_greater() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::Greater,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[0, 0, 1]));
}

#[test]
fn execute_greater_eq() {
    let mut machine = Machine::new(3, input(&[&[10, 0, 10], &[10, 10, 0]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::BinaryOp {
        op: BinaryOp::GreaterEq,
        lhs: 0,
        rhs: 1,
    });

    assert_eq!(machine.output(), output(&[1, 0, 1]));
}

#[test]
fn execute_if() {
    let mut machine = Machine::new(2, input(&[&[0, 1], &[4, 5], &[7, 8]]));

    machine.exec(Instruction::Input(0));
    machine.exec(Instruction::Input(1));
    machine.exec(Instruction::Input(2));
    machine.exec(Instruction::If {
        cond: 0,
        r#true: 1,
        r#false: 2,
    });

    assert_eq!(machine.output(), output(&[7, 5]));
}

fn input(value: &[&[Value]]) -> Box<[Box<[Value]>]> {
    value.iter().copied().map(Box::from).collect()
}

fn output(value: &[Value]) -> Vector {
    value.iter().copied().collect()
}
