use std::{
    iter::{repeat, Repeat, Take},
    ops::Index,
    slice,
    sync::Arc,
};

pub type Value = i64;

#[derive(Debug, Clone)]
pub enum Vector {
    Array(Arc<Box<[Value]>>),
    Const { value: Value, len: usize },
}

impl Vector {
    pub fn iter(&self) -> VectorIter {
        match self {
            Vector::Array(array) => VectorIter::Array(array.iter()),
            Vector::Const { value, len } => VectorIter::Const(repeat(value).take(*len)),
        }
    }
}

impl From<Box<[Value]>> for Vector {
    fn from(value: Box<[Value]>) -> Self {
        Self::Array(Arc::new(value))
    }
}

impl FromIterator<Value> for Vector {
    fn from_iter<T: IntoIterator<Item = Value>>(iter: T) -> Self {
        Self::Array(Arc::new(iter.into_iter().collect()))
    }
}

impl Index<usize> for Vector {
    type Output = Value;

    fn index(&self, index: usize) -> &Self::Output {
        match self {
            Vector::Array(array) => &array[index],
            Vector::Const { value, .. } => value,
        }
    }
}

impl PartialEq for Vector {
    fn eq(&self, other: &Self) -> bool {
        self.iter().eq(other.iter())
    }
}

impl Eq for Vector {}

#[derive(Debug, Clone)]
pub enum VectorIter<'a> {
    Array(slice::Iter<'a, Value>),
    Const(Take<Repeat<&'a Value>>),
}

impl<'a> Iterator for VectorIter<'a> {
    type Item = &'a Value;

    fn next(&mut self) -> Option<Self::Item> {
        match self {
            VectorIter::Array(i) => i.next(),
            VectorIter::Const(i) => i.next(),
        }
    }
}
